#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:39 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

from .generator import Generator
from ..toolkit.utility import _mergeFuel as merge
from ..toolkit.utility import _fillFuel as fill
from ..toolkit.utility import _deductFuel as deduct

__all__ = ('Generator', 'depot', 'merge', 'fill', 'deduct')